@@ -11,7 +11,8 @@
 
 """
 @package gribapi
-@brief This package is the \b Python 3 interface to ecCodes. It offers almost one to one bindings to the C API functions.
+@brief This package is the \b Python 3 interface to ecCodes.
+It offers almost one to one bindings to the C API functions.
 
 The Python 3 interface to ecCodes uses the <a href="http://numpy.scipy.org/"><b>NumPy</b></a> package
 as the container of choice for the possible arrays of values that can be encoded/decoded in and from a message.
@@ -19,19 +20,23 @@ Numpy is a package used for scientific computing in Python and an efficient cont
 
 @em Requirements:
 
-    - Python 3.5 or higher
+    - Python 3.8 or higher
     - NumPy
 
 """
 
-from functools import wraps
-import sys
 import os
+import sys
+from functools import wraps
+
 import numpy as np
 
-from .bindings import ENC, ffi, lib
-from .bindings import __version__ as bindings_version  # noqa
+from gribapi.errors import GribInternalError
+
 from . import errors
+from .bindings import ENC
+from .bindings import __version__ as bindings_version  # noqa
+from .bindings import ffi, lib, library_path
 
 try:
     type(file)
@@ -41,11 +46,7 @@ except NameError:
     file = io.IOBase
     long = int
 
-KEYTYPES = {
-    1: int,
-    2: float,
-    3: str,
-}
+KEYTYPES = {1: int, 2: float, 3: str, 4: bytes}
 
 CODES_PRODUCT_ANY = 0
 """ Generic product kind """
@@ -64,6 +65,10 @@ CODES_PRODUCT_TAF = 5
 GRIB_MISSING_DOUBLE = -1e100
 GRIB_MISSING_LONG = 2147483647
 
+# Constants for GRIB nearest neighbour
+GRIB_NEAREST_SAME_GRID = 1 << 0
+GRIB_NEAREST_SAME_DATA = 1 << 1
+GRIB_NEAREST_SAME_POINT = 1 << 2
 
 # ECC-1029: Disable function-arguments type-checking unless
 # environment variable is defined and equal to 1
@@ -93,13 +98,12 @@ def require(**_params_):
                 param = kw[name]
                 if isinstance(allowed_types, type):
                     allowed_types = (allowed_types,)
-                assert any([isinstance(param, type1) for type1 in allowed_types]), (
-                    "Parameter '%s' should be of type %s (instead of %s)"
-                    % (
-                        name,
-                        " or ".join([t.__name__ for t in allowed_types]),
-                        type(param).__name__,
-                    )
+                assert any(
+                    [isinstance(param, type1) for type1 in allowed_types]
+                ), "Parameter '%s' should be of type %s (instead of %s)" % (
+                    name,
+                    " or ".join([t.__name__ for t in allowed_types]),
+                    type(param).__name__,
                 )
             return _func_(**kw)
 
@@ -156,21 +160,19 @@ def err_last(func):
 
 
 def get_handle(msgid):
-    assert isinstance(msgid, int)
     h = ffi.cast("grib_handle*", msgid)
     if h == ffi.NULL:
-        raise errors.InvalidGribError(f"get_handle: Bad message ID {msgid}")
+        raise errors.NullHandleError(f"get_handle: Bad message ID {msgid}")
     return h
 
 
 def put_handle(handle):
     if handle == ffi.NULL:
-        raise errors.InvalidGribError(f"put_handle: Bad message ID {handle}")
+        raise errors.NullHandleError("put_handle: Bad message ID (handle is NULL)")
     return int(ffi.cast("size_t", handle))
 
 
 def get_multi_handle(msgid):
-    assert isinstance(msgid, int)
     return ffi.cast("grib_multi_handle*", msgid)
 
 
@@ -179,7 +181,6 @@ def put_multi_handle(handle):
 
 
 def get_index(indexid):
-    assert isinstance(indexid, int)
     return ffi.cast("grib_index*", indexid)
 
 
@@ -188,7 +189,6 @@ def put_index(indexh):
 
 
 def get_iterator(iterid):
-    assert isinstance(iterid, int)
     return ffi.cast("grib_iterator*", iterid)
 
 
@@ -197,7 +197,6 @@ def put_iterator(iterh):
 
 
 def get_grib_keys_iterator(iterid):
-    assert isinstance(iterid, int)
     return ffi.cast("grib_keys_iterator*", iterid)
 
 
@@ -206,7 +205,6 @@ def put_grib_keys_iterator(iterh):
 
 
 def get_bufr_keys_iterator(iterid):
-    assert isinstance(iterid, int)
     return ffi.cast("bufr_keys_iterator*", iterid)
 
 
@@ -244,7 +242,7 @@ def gts_new_from_file(fileobj, headers_only=False):
     @return               id of the GTS loaded in memory or None
     @exception CodesInternalError
     """
-    # err, h = err_last(lib.gts_new_from_file)(ffi.NULL, fileobj)
+
     err, h = err_last(lib.codes_handle_new_from_file)(
         ffi.NULL, fileobj, CODES_PRODUCT_GTS
     )
@@ -273,7 +271,7 @@ def metar_new_from_file(fileobj, headers_only=False):
     @return               id of the METAR loaded in memory or None
     @exception CodesInternalError
     """
-    # err, h = err_last(lib.metar_new_from_file)(ffi.NULL, fileobj)
+
     err, h = err_last(lib.codes_handle_new_from_file)(
         ffi.NULL, fileobj, CODES_PRODUCT_METAR
     )
@@ -315,7 +313,7 @@ def codes_new_from_file(fileobj, product_kind, headers_only=False):
         return gts_new_from_file(fileobj, headers_only)
     if product_kind == CODES_PRODUCT_ANY:
         return any_new_from_file(fileobj, headers_only)
-    raise Exception("Invalid product kind: " + product_kind)
+    raise ValueError("Invalid product kind %d" % product_kind)
 
 
 @require(fileobj=file)
@@ -400,7 +398,7 @@ def grib_new_from_file(fileobj, headers_only=False):
     @return               id of the grib loaded in memory or None
     @exception CodesInternalError
     """
-    # err, h = err_last(lib.grib_new_from_file)(ffi.NULL, fileobj, headers_only)
+
     err, h = err_last(lib.codes_handle_new_from_file)(
         ffi.NULL, fileobj, CODES_PRODUCT_GRIB
     )
@@ -457,7 +455,7 @@ def grib_multi_support_reset_file(fileobj):
     @brief Reset file handle in multiple field support mode
     """
     context = lib.grib_context_get_default()
-    GRIB_CHECK(lib.grib_multi_support_reset_file(context, fileobj))
+    lib.grib_multi_support_reset_file(context, fileobj)
 
 
 @require(msgid=int)
@@ -491,7 +489,7 @@ def grib_get_string(msgid, key):
     length_p = ffi.new("size_t *", length)
     err = lib.grib_get_string(h, key.encode(ENC), values, length_p)
     GRIB_CHECK(err)
-    return ffi.string(values, length_p[0]).decode(ENC)
+    return _decode_bytes(values, length_p[0])
 
 
 @require(msgid=int, key=str, value=str)
@@ -754,12 +752,10 @@ def grib_iterator_next(iterid):
     lat_p = ffi.new("double*")
     lon_p = ffi.new("double*")
     value_p = ffi.new("double*")
-    err = lib.grib_iterator_next(iterh, lat_p, lon_p, value_p)
-    if err == 0:
+    retval = lib.grib_iterator_next(iterh, lat_p, lon_p, value_p)
+    if retval == 0:
+        # No more data available. End of iteration
         return []
-    elif err < 0:
-        GRIB_CHECK(err)
-        return None
     else:
         return (lat_p[0], lon_p[0], value_p[0])
 
@@ -805,8 +801,7 @@ def grib_keys_iterator_next(iterid):
     """
     kih = get_grib_keys_iterator(iterid)
     res = lib.grib_keys_iterator_next(kih)
-    if res < 0:
-        GRIB_CHECK(res)
+    # res is 0 or 1
     return res
 
 
@@ -820,7 +815,6 @@ def grib_keys_iterator_delete(iterid):
     @param iterid      keys iterator id created with @ref grib_keys_iterator_new
     @exception CodesInternalError
     """
-    # aa: THIS LEAKS MEMORY as it doesn't free all the connected iterators
     kih = get_grib_keys_iterator(iterid)
     lib.grib_keys_iterator_delete(kih)
 
@@ -836,7 +830,6 @@ def grib_keys_iterator_get_name(iterid):
     @return          key name to be retrieved
     @exception CodesInternalError
     """
-    # aa: missing call to grib_keys_iterator_get_accessor
     kih = get_grib_keys_iterator(iterid)
     name = lib.grib_keys_iterator_get_name(kih)
     return ffi.string(name).decode(ENC)
@@ -891,8 +884,7 @@ def codes_bufr_keys_iterator_next(iterid):
     """
     bki = get_bufr_keys_iterator(iterid)
     res = lib.codes_bufr_keys_iterator_next(bki)
-    if res < 0:
-        GRIB_CHECK(res)
+    # res is 0 or 1
     return res
 
 
@@ -972,7 +964,7 @@ def grib_get_double(msgid, key):
     return value_p[0]
 
 
-@require(msgid=int, key=str, value=(int, float, str))
+@require(msgid=int, key=str, value=(int, float, np.float16, np.float32, np.int64, str))
 def grib_set_long(msgid, key, value):
     """
     @brief Set the integer value for a key in a message.
@@ -991,13 +983,13 @@ def grib_set_long(msgid, key, value):
         raise TypeError("Invalid type")
 
     if value > sys.maxsize:
-        raise TypeError("Invalid type")
+        raise ValueError("Value too large")
 
     h = get_handle(msgid)
     GRIB_CHECK(lib.grib_set_long(h, key.encode(ENC), value))
 
 
-@require(msgid=int, key=str, value=(int, float, str))
+@require(msgid=int, key=str, value=(int, float, np.float16, np.float32, str))
 def grib_set_double(msgid, key, value):
     """
     @brief Set the double value for a key in a message.
@@ -1038,7 +1030,9 @@ def codes_new_from_samples(samplename, product_kind):
         return grib_new_from_samples(samplename)
     if product_kind == CODES_PRODUCT_BUFR:
         return codes_bufr_new_from_samples(samplename)
-    raise Exception("Invalid product kind: " + product_kind)
+    if product_kind == CODES_PRODUCT_ANY:
+        return codes_any_new_from_samples(samplename)
+    raise ValueError("Invalid product kind %d" % product_kind)
 
 
 @require(samplename=str)
@@ -1083,6 +1077,25 @@ def codes_bufr_new_from_samples(samplename):
     return put_handle(h)
 
 
+@require(samplename=str)
+def codes_any_new_from_samples(samplename):
+    """
+    @brief Create a new valid message from a sample.
+
+    The available samples are picked up from the directory pointed to
+    by the environment variable ECCODES_SAMPLES_PATH.
+    To know where the samples directory is run the codes_info tool.\n
+
+    @param samplename   name of the sample to be used
+    @return             id of the message loaded in memory
+    @exception CodesInternalError
+    """
+    h = lib.codes_handle_new_from_samples(ffi.NULL, samplename.encode(ENC))
+    if h == ffi.NULL:
+        raise errors.FileNotFoundError(f"any_new_from_samples failed: {samplename}")
+    return put_handle(h)
+
+
 @require(msgid_src=int, msgid_dst=int)
 def codes_bufr_copy_data(msgid_src, msgid_dst):
     """
@@ -1104,23 +1117,29 @@ def codes_bufr_copy_data(msgid_src, msgid_dst):
 
 
 @require(msgid_src=int)
-def grib_clone(msgid_src):
+def grib_clone(msgid_src, headers_only=False):
     r"""
     @brief Create a copy of a message.
 
     Create a copy of a given message (\em msgid_src) resulting in a new
     message in memory (\em msgid_dest) identical to the original one.
+    If the headers_only option is enabled, the clone will not contain
+    the Bitmap and Data sections
 
     \b Examples: \ref grib_clone.py "grib_clone.py"
 
-    @param msgid_src   id of message to be cloned
-    @return            id of clone
+    @param msgid_src    id of message to be cloned
+    @param headers_only whether or not to clone the message with the headers only
+    @return             id of clone
     @exception CodesInternalError
     """
     h_src = get_handle(msgid_src)
-    h_dest = lib.grib_handle_clone(h_src)
+    if headers_only:
+        h_dest = lib.grib_handle_clone_headers_only(h_src)
+    else:
+        h_dest = lib.grib_handle_clone(h_src)
     if h_dest == ffi.NULL:
-        raise errors.InvalidGribError("clone failed")
+        raise errors.MessageInvalidError("clone failed")
     return put_handle(h_dest)
 
 
@@ -1142,6 +1161,8 @@ def grib_set_double_array(msgid, key, inarray):
     length = len(inarray)
     if isinstance(inarray, np.ndarray):
         nd = inarray
+        # ECC-1555
+        length = inarray.size
         if length > 0:
             if not isinstance(nd[0], float):
                 # ECC-1042: input array of integers
@@ -1176,6 +1197,39 @@ def grib_get_double_array(msgid, key):
     return arr
 
 
+@require(msgid=int, key=str)
+def grib_get_float_array(msgid, key):
+    """
+    @brief Get the value of the key as a NumPy array of floats.
+
+    @param msgid   id of the message loaded in memory
+    @param key     key name
+    @return        numpy.ndarray
+    @exception CodesInternalError
+    """
+    h = get_handle(msgid)
+    nval = grib_get_size(msgid, key)
+    length_p = ffi.new("size_t*", nval)
+    arr = np.empty((nval,), dtype="float32")
+    vals_p = ffi.cast("float *", arr.ctypes.data)
+    err = lib.grib_get_float_array(h, key.encode(ENC), vals_p, length_p)
+    GRIB_CHECK(err)
+    return arr
+
+
+# See ECC-1246
+def _decode_bytes(binput, maxlen=None):
+    if maxlen:
+        a_str = ffi.string(binput, maxlen)
+    else:
+        a_str = ffi.string(binput)
+    # Check for a MISSING value i.e., each character has all its bits=1
+    if all(x == 255 for x in a_str):
+        return ""
+    # Replace with a suitable replacement character rather than throw an exception
+    return a_str.decode(ENC, "replace")
+
+
 @require(msgid=int, key=str)
 def grib_get_string_array(msgid, key):
     """
@@ -1194,7 +1248,7 @@ def grib_get_string_array(msgid, key):
     size_p = ffi.new("size_t *", size)
     err = lib.grib_get_string_array(h, key.encode(ENC), values, size_p)
     GRIB_CHECK(err)
-    return [ffi.string(values[i]).decode(ENC) for i in range(size_p[0])]
+    return [_decode_bytes(values[i]) for i in range(size_p[0])]
 
 
 @require(msgid=int, key=str)
@@ -1320,7 +1374,10 @@ def grib_index_new_from_file(filename, keys):
 
     @param filename   path of the file to index on
     @param keys       sequence of keys to index on.
-                      The type of the key can be explicitly declared appending :l for long (or alternatively :i), :d for double, :s for string to the key name.
+                      The type of the key can be explicitly declared appending
+                      :l for long (or alternatively :i),
+                      :d for double,
+                      :s for string to the key name.
     @return           index id
     @exception CodesInternalError
     """
@@ -1388,12 +1445,13 @@ def grib_index_get_long(indexid, key):
     @brief Get the distinct values of the key in argument contained in the index.
     The key must belong to the index.
 
-    This function is used when the type of the key was explicitly defined as long or when the native type of the key is long.
+    This function is used when the type of the key was explicitly defined as long or when the native type of
+    the key is long.
 
     \b Examples: \ref grib_index.py "grib_index.py"
 
     @param indexid   id of an index created from a file. The index must have been created with the key in argument.
-    @param key       key for wich the values are returned
+    @param key       key for which the values are returned
     @return          tuple with values of key in index
     @exception CodesInternalError
     """
@@ -1413,12 +1471,13 @@ def grib_index_get_string(indexid, key):
     @brief Get the distinct values of the key in argument contained in the index.
     The key must belong to the index.
 
-    This function is used when the type of the key was explicitly defined as string or when the native type of the key is string.
+    This function is used when the type of the key was explicitly defined as string or when the native type of
+    the key is string.
 
     \b Examples: \ref grib_index.py "grib_index.py"
 
     @param indexid   id of an index created from a file. The index must have been created with the key in argument.
-    @param key       key for wich the values are returned
+    @param key       key for which the values are returned
     @return          tuple with values of key in index
     @exception CodesInternalError
     """
@@ -1439,12 +1498,13 @@ def grib_index_get_double(indexid, key):
     @brief Get the distinct values of the key in argument contained in the index.
     The key must belong to the index.
 
-    This function is used when the type of the key was explicitly defined as double or when the native type of the key is double.
+    This function is used when the type of the key was explicitly defined as double or when the native type
+    of the key is double.
 
     \b Examples: \ref grib_index.py "grib_index.py"
 
     @param indexid  id of an index created from a file. The index must have been created with the key in argument.
-    @param key      key for wich the values are returned
+    @param key      key for which the values are returned
     @return         tuple with values of key in index
     @exception CodesInternalError
     """
@@ -1453,9 +1513,9 @@ def grib_index_get_double(indexid, key):
 
     values_p = ffi.new("double[]", nval)
     size_p = ffi.new("size_t *", nval)
-    err = lib.grib_index_get_doule(ih, key.encode(ENC), values_p, size_p)
+    err = lib.grib_index_get_double(ih, key.encode(ENC), values_p, size_p)
     GRIB_CHECK(err)
-    return tuple(int(values_p[i]) for i in range(size_p[0]))
+    return tuple(values_p[i] for i in range(size_p[0]))
 
 
 @require(indexid=int, key=str, value=int)
@@ -1464,7 +1524,8 @@ def grib_index_select_long(indexid, key, value):
     @brief Select the message subset with key==value.
     The value is an integer.
 
-    The key must have been created with integer type or have integer as native type if the type was not explicitly defined in the index creation.
+    The key must have been created with integer type or have integer as native type if the type
+    was not explicitly defined in the index creation.
 
     \b Examples: \ref grib_index.py "grib_index.py"
 
@@ -1483,7 +1544,8 @@ def grib_index_select_double(indexid, key, value):
     @brief Select the message subset with key==value.
     The value is a double.
 
-    The key must have been created with integer type or have integer as native type if the type was not explicitly defined in the index creation.
+    The key must have been created with integer type or have integer as native type if the type was
+    not explicitly defined in the index creation.
 
     \b Examples: \ref grib_index.py "grib_index.py"
 
@@ -1502,7 +1564,8 @@ def grib_index_select_string(indexid, key, value):
     @brief Select the message subset with key==value.
     The value is an integer.
 
-    The key must have been created with string type or have string as native type if the type was not explicitly defined in the index creation.
+    The key must have been created with string type or have string as native type if the type
+    was not explicitly defined in the index creation.
 
     \b Examples: \ref grib_index.py "grib_index.py"
 
@@ -1521,7 +1584,8 @@ def grib_new_from_index(indexid):
     @brief Create a new handle from an index after having selected the key values.
 
     All the keys belonging to the index must be selected before calling this function.
-    Successive calls to this function will return all the handles compatible with the constraints defined selecting the values of the index keys.
+    Successive calls to this function will return all the handles compatible with the constraints
+    defined selecting the values of the index keys.
 
     The message can be accessed through its gribid and will be available until @ref grib_release is called.
 
@@ -1589,7 +1653,7 @@ def grib_get_double_element(msgid, key, index):
     """
     h = get_handle(msgid)
     value_p = ffi.new("double*")
-    err = lib.grib_get_double_element(h, key.encode(ENC), index, value_p)  # TODO
+    err = lib.grib_get_double_element(h, key.encode(ENC), index, value_p)
     GRIB_CHECK(err)
     return value_p[0]
 
@@ -1675,9 +1739,7 @@ def grib_set_key_vals(gribid, key_vals):
             if not isinstance(kv, str):
                 raise TypeError("Invalid list/tuple element type '%s'" % kv)
             if "=" not in str(kv):
-                raise errors.GribInternalError(
-                    "Invalid list/tuple element format '%s'" % kv
-                )
+                raise GribInternalError("Invalid list/tuple element format '%s'" % kv)
             if len(key_vals_str) > 0:
                 key_vals_str += ","
             key_vals_str += kv
@@ -1823,7 +1885,8 @@ def grib_find_nearest(gribid, inlat, inlon, is_lsm=False, npoints=1):
 @require(gribid=int, is_lsm=bool)
 def grib_find_nearest_multiple(gribid, is_lsm, inlats, inlons):
     """
-    @brief Find the nearest point of a set of points whose latitudes and longitudes are given in the inlats, inlons arrays respectively
+    @brief Find the nearest point of a set of points whose latitudes and longitudes are given in
+    the inlats, inlons arrays respectively
 
     @param gribid     id of the GRIB message loaded in memory
     @param is_lsm     True if the nearest land point is required otherwise False.
@@ -1882,7 +1945,7 @@ def grib_get_native_type(msgid, key):
     """
     @brief Retrieve the native type of a key.
 
-    Possible values can be int, float or string.
+    Possible values can be int, float or str.
 
     @param msgid   id of the message loaded in memory
     @param key     key we want to find out the type for
@@ -1906,7 +1969,7 @@ def grib_get(msgid, key, ktype=None):
 
     The type of value returned depends on the native type of the requested key.
     The type of value returned can be forced by using the type argument of the
-    function. The type argument can be int, float or str.
+    function. The ktype argument can be int, float, str or bytes.
 
     The \em msgid references a message loaded in memory.
 
@@ -1916,7 +1979,7 @@ def grib_get(msgid, key, ktype=None):
 
     @param msgid      id of the message loaded in memory
     @param key        key name
-    @param ktype      the type we want the output in (int, float or str), native type if not specified
+    @param ktype      the type we want the output in, native type if not specified
     @return           scalar value of key as int, float or str
     @exception CodesInternalError
     """
@@ -1933,6 +1996,8 @@ def grib_get(msgid, key, ktype=None):
         result = grib_get_double(msgid, key)
     elif ktype is str:
         result = grib_get_string(msgid, key)
+    elif ktype is bytes:
+        result = grib_get_string(msgid, key)
 
     return result
 
@@ -1944,13 +2009,13 @@ def grib_get_array(msgid, key, ktype=None):
 
     The type of the array returned depends on the native type of the requested key.
     For numeric data, the output array will be stored in a NumPy ndarray.
-    The type of value returned can be forced by using the type argument of the function.
-    The type argument can be int, float or string.
+    The type of value returned can be forced by using the ktype argument of the function.
+    The ktype argument can be int, float, float32, float64, str or bytes.
 
-    @param msgid      id of the message loaded in memory
-    @param key        the key to get the value for
-    @param ktype      the type we want the output in (can be int, float or string), native type if not specified
-    @return           numpy.ndarray
+    @param msgid  id of the message loaded in memory
+    @param key    the key to get the value for
+    @param ktype  the type we want the output in, native type if not specified
+    @return       numpy.ndarray or None
     @exception CodesInternalError
     """
     if ktype is None:
@@ -1959,16 +2024,20 @@ def grib_get_array(msgid, key, ktype=None):
     result = None
     if ktype is int:
         result = grib_get_long_array(msgid, key)
-    elif ktype is float:
+    elif ktype is float or ktype is np.float64:
         result = grib_get_double_array(msgid, key)
+    elif ktype is np.float32:
+        result = grib_get_float_array(msgid, key)
     elif ktype is str:
         result = grib_get_string_array(msgid, key)
+    elif ktype is bytes:
+        result = grib_get_string_array(msgid, key)
 
     return result
 
 
 @require(gribid=int)
-def grib_get_values(gribid):
+def grib_get_values(gribid, ktype=float):
     """
     @brief Retrieve the contents of the 'values' key for a GRIB message.
 
@@ -1976,11 +2045,23 @@ def grib_get_values(gribid):
 
     \b Examples: \ref grib_print_data.py "grib_print_data.py", \ref grib_samples.py "grib_samples.py"
 
-    @param gribid   id of the GRIB loaded in memory
-    @return         numpy.ndarray
+    @param gribid    id of the GRIB loaded in memory
+    @param ktype     data type of the result: numpy.float32 or numpy.float64
+    @return          numpy.ndarray
     @exception CodesInternalError
     """
-    return grib_get_double_array(gribid, "values")
+    result = None
+
+    if ktype is np.float32:
+        result = grib_get_float_array(gribid, "values")
+    elif ktype is np.float64 or ktype is float:
+        result = grib_get_double_array(gribid, "values")
+    else:
+        raise TypeError(
+            f"Unsupported data type {ktype}. Supported data types are numpy.float32 and numpy.float64"
+        )
+
+    return result
 
 
 @require(gribid=int)
@@ -2038,9 +2119,9 @@ def grib_set(msgid, key, value):
     @param value      scalar value to set for key
     @exception CodesInternalError
     """
-    if isinstance(value, int):
+    if isinstance(value, (int, np.int64)):
         grib_set_long(msgid, key, value)
-    elif isinstance(value, float):
+    elif isinstance(value, (float, np.float16, np.float32, np.float64)):
         grib_set_double(msgid, key, value)
     elif isinstance(value, str):
         grib_set_string(msgid, key, value)
@@ -2048,8 +2129,11 @@ def grib_set(msgid, key, value):
     #    # The value passed in is iterable; i.e. a list or array etc
     #    grib_set_array(msgid, key, value)
     else:
-        raise errors.GribInternalError(
-            "Invalid type of value when setting key '%s'." % key
+        hint = ""
+        if hasattr(value, "__iter__"):
+            hint = " (Hint: for array keys use codes_set_array(msgid, key, value))"
+        raise GribInternalError(
+            "Invalid type of value when setting key '%s'%s." % (key, hint)
         )
 
 
@@ -2076,16 +2160,15 @@ def grib_set_array(msgid, key, value):
     except TypeError:
         pass
 
-    if isinstance(val0, float):
+    if isinstance(val0, (float, np.float16, np.float32, np.float64)):
         grib_set_double_array(msgid, key, value)
     elif isinstance(val0, str):
         grib_set_string_array(msgid, key, value)
     else:
-        # Note: Cannot do isinstance(val0,int) for numpy.int64
         try:
             int(val0)
         except (ValueError, TypeError):
-            raise errors.GribInternalError(
+            raise GribInternalError(
                 "Invalid type of value when setting key '%s'." % key
             )
         grib_set_long_array(msgid, key, value)
@@ -2142,9 +2225,7 @@ def grib_index_select(indexid, key, value):
     elif isinstance(value, str):
         grib_index_select_string(indexid, key, value)
     else:
-        raise errors.GribInternalError(
-            "Invalid type of value when setting key '%s'." % key
-        )
+        raise GribInternalError("Invalid type of value when setting key '%s'." % key)
 
 
 @require(indexid=int, filename=str)
@@ -2204,11 +2285,12 @@ def grib_gts_header(flag):
         lib.grib_gts_header_off(context)
 
 
-def grib_get_api_version():
+def grib_get_api_version(vformat=str):
     """
     @brief Get the API version.
 
-    Returns the version of the API as a string in the format "major.minor.revision".
+    Returns the version of the API as a string in the format "major.minor.revision"
+    or as an integer (10000*major + 100*minor + revision)
     """
 
     def div(v, d):
@@ -2218,11 +2300,14 @@ def grib_get_api_version():
         raise RuntimeError("Could not load the ecCodes library!")
 
     v = lib.grib_get_api_version()
-    v, revision = div(v, 100)
-    v, minor = div(v, 100)
-    major = v
 
-    return "%d.%d.%d" % (major, minor, revision)
+    if vformat is str:
+        v, revision = div(v, 100)
+        v, minor = div(v, 100)
+        major = v
+        return "%d.%d.%d" % (major, minor, revision)
+    else:
+        return v
 
 
 __version__ = grib_get_api_version()
@@ -2240,6 +2325,21 @@ def codes_get_version_info():
     return vinfo
 
 
+@require(order=int)
+def codes_get_gaussian_latitudes(order):
+    """
+    @brief Return the Gaussian latitudes
+
+    @param order    The Gaussian order/number (also called the truncation)
+    @return         A list of latitudes with 2*order elements
+    """
+    num_elems = 2 * order
+    outlats_p = ffi.new("double[]", num_elems)
+    err = lib.grib_get_gaussian_latitudes(order, outlats_p)
+    GRIB_CHECK(err)
+    return outlats_p
+
+
 @require(msgid=int)
 def grib_get_message(msgid):
     """
@@ -2283,7 +2383,7 @@ def grib_new_from_message(message):
         message = message.encode(ENC)
     h = lib.grib_handle_new_from_message_copy(ffi.NULL, message, len(message))
     if h == ffi.NULL:
-        raise errors.InvalidGribError("new_from_message failed")
+        raise errors.MessageInvalidError("new_from_message failed")
     return put_handle(h)
 
 
@@ -2327,6 +2427,13 @@ def grib_set_samples_path(samples_path):
     lib.grib_context_set_samples_path(context, samples_path.encode(ENC))
 
 
+def grib_context_delete():
+    """
+    @brief Wipe all the cached data and definitions files in the context
+    """
+    lib.grib_context_delete(ffi.NULL)
+
+
 def codes_bufr_multi_element_constant_arrays_on():
     """
     @brief BUFR: Turn on the mode where you get multiple elements
@@ -2349,6 +2456,19 @@ def codes_bufr_multi_element_constant_arrays_off():
     lib.codes_bufr_multi_element_constant_arrays_off(context)
 
 
+@require(msgid=int)
+def codes_dump(msgid, output_fileobj=sys.stdout, mode="wmo", flags=0):
+    """
+    @brief Print all keys to an output file object, with the given dump mode and flags
+
+    @param msgid          id of the message loaded in memory
+    @param output_fileobj output file object e.g., sys.stdout
+    @param mode           dump mode e.g., "wmo", "debug", "json"
+    """
+    h = get_handle(msgid)
+    lib.grib_dump_content(h, output_fileobj, mode.encode(ENC), flags, ffi.NULL)
+
+
 # Convert the C codes_bufr_header struct to a Python dictionary
 def _convert_struct_to_dict(s):
     result = {}
@@ -2364,7 +2484,7 @@ def _convert_struct_to_dict(s):
 
 def codes_bufr_extract_headers(filepath, is_strict=True):
     """
-    @brief BUFR header extraction (EXPERIMENTAL FEATURE)
+    @brief BUFR header extraction
 
     @param filepath       path of input BUFR file
     @param is_strict      fail as soon as any invalid BUFR message is encountered
@@ -2393,3 +2513,186 @@ def codes_bufr_extract_headers(filepath, is_strict=True):
     while i < num_messages:
         yield _convert_struct_to_dict(headers[i])
         i += 1
+
+
+@require(msgid=int)
+def codes_bufr_key_is_header(msgid, key):
+    """
+    @brief Check if the BUFR key is in the header or in the data section.
+
+    If the data section has not been unpacked, then passing in a key from
+    the data section will throw KeyValueNotFoundError.
+
+    @param msgid      id of the BUFR message loaded in memory
+    @param key        key name
+    @return           1->header, 0->data section
+    @exception CodesInternalError
+    """
+    h = get_handle(msgid)
+    err, value = err_last(lib.codes_bufr_key_is_header)(h, key.encode(ENC))
+    GRIB_CHECK(err)
+    return value
+
+
+@require(msgid=int)
+def codes_bufr_key_is_coordinate(msgid, key):
+    """
+    @brief Check if the BUFR key corresponds to a coordinate descriptor.
+
+    If the data section has not been unpacked, then passing in a key from
+    the data section will throw KeyValueNotFoundError.
+
+    @param msgid      id of the BUFR message loaded in memory
+    @param key        key name
+    @return           1->coordinate, 0->not coordinate
+    @exception CodesInternalError
+    """
+    h = get_handle(msgid)
+    err, value = err_last(lib.codes_bufr_key_is_coordinate)(h, key.encode(ENC))
+    GRIB_CHECK(err)
+    return value
+
+
+def codes_extract_offsets(filepath, product_kind, is_strict=True):
+    """
+    @brief Message offset extraction
+
+    @param filepath      path of input file
+    @param product_kind  one of CODES_PRODUCT_GRIB, CODES_PRODUCT_BUFR, CODES_PRODUCT_ANY or CODES_PRODUCT_GTS
+    @param is_strict     if True, fail as soon as any invalid message is encountered
+    @return              a generator that yields offsets (as integers)
+    @exception CodesInternalError
+    """
+    context = lib.grib_context_get_default()
+    offsets_p = ffi.new("long int**")
+    num_message_p = ffi.new("int*")
+
+    err = lib.codes_extract_offsets_malloc(
+        context, filepath.encode(ENC), product_kind, offsets_p, num_message_p, is_strict
+    )
+    GRIB_CHECK(err)
+
+    num_messages = num_message_p[0]
+    offsets = offsets_p[0]
+
+    i = 0
+    while i < num_messages:
+        yield offsets[i]
+        i += 1
+
+
+def codes_extract_offsets_sizes(filepath, product_kind, is_strict=True):
+    """
+    @brief Message offset and size extraction
+
+    @param filepath      path of input file
+    @param product_kind  one of CODES_PRODUCT_GRIB, CODES_PRODUCT_BUFR, CODES_PRODUCT_ANY or CODES_PRODUCT_GTS
+    @param is_strict     if True, fail as soon as any invalid message is encountered
+    @return              a generator that yields lists of pairs of offsets and sizes (as integers)
+    @exception CodesInternalError
+    """
+    context = lib.grib_context_get_default()
+    offsets_p = ffi.new("long int**")
+    sizes_p = ffi.new("size_t**")
+    num_message_p = ffi.new("int*")
+
+    err = lib.codes_extract_offsets_sizes_malloc(
+        context,
+        filepath.encode(ENC),
+        product_kind,
+        offsets_p,
+        sizes_p,
+        num_message_p,
+        is_strict,
+    )
+    GRIB_CHECK(err)
+
+    num_messages = num_message_p[0]
+    offsets = offsets_p[0]
+    sizes = sizes_p[0]
+
+    i = 0
+    while i < num_messages:
+        yield (offsets[i], sizes[i])
+        i += 1
+
+
+# -------------------------------
+# EXPERIMENTAL FEATURES
+# -------------------------------
+@require(msgid=int)
+def grib_nearest_new(msgid):
+    h = get_handle(msgid)
+    err, nid = err_last(lib.grib_nearest_new)(h)
+    GRIB_CHECK(err)
+    return put_grib_nearest(nid)
+
+
+def put_grib_nearest(nid):
+    return int(ffi.cast("size_t", nid))
+
+
+def get_grib_nearest(nid):
+    return ffi.cast("grib_nearest*", nid)
+
+
+@require(nid=int)
+def grib_nearest_delete(nid):
+    nh = get_grib_nearest(nid)
+    lib.grib_nearest_delete(nh)
+
+
+@require(nid=int, gribid=int)
+def grib_nearest_find(nid, gribid, inlat, inlon, flags, is_lsm=False, npoints=4):
+    # flags has to be one of:
+    #  GRIB_NEAREST_SAME_GRID
+    #  GRIB_NEAREST_SAME_DATA
+    #  GRIB_NEAREST_SAME_POINT
+    if npoints != 4:
+        raise errors.FunctionNotImplementedError(
+            "grib_nearest_find npoints argument: Only 4 points supported"
+        )
+    if is_lsm:
+        raise errors.FunctionNotImplementedError(
+            "grib_nearest_find is_lsm argument: Land sea mask not supported"
+        )
+
+    h = get_handle(gribid)
+    outlats_p = ffi.new("double[]", npoints)
+    outlons_p = ffi.new("double[]", npoints)
+    values_p = ffi.new("double[]", npoints)
+    distances_p = ffi.new("double[]", npoints)
+    indexes_p = ffi.new("int[]", npoints)
+    size = ffi.new("size_t *")
+    nh = get_grib_nearest(nid)
+    err = lib.grib_nearest_find(
+        nh,
+        h,
+        inlat,
+        inlon,
+        flags,
+        outlats_p,
+        outlons_p,
+        values_p,
+        distances_p,
+        indexes_p,
+        size,
+    )
+    GRIB_CHECK(err)
+    result = []
+    for i in range(npoints):
+        result.append(
+            Bunch(
+                lat=outlats_p[i],
+                lon=outlons_p[i],
+                value=values_p[i],
+                distance=distances_p[i],
+                index=indexes_p[i],
+            )
+        )
+
+    return tuple(result)
+
+
+def codes_get_library_path():
+    return library_path
